#include "module_traffic_light_manager.h"
#include "db_api/CDBSingletonDefine.h"
#include <db_api/CDBResultSet.h>

traffic_light_manager* traffic_light_manager::instance()
{
    static traffic_light_manager tlm;
    return &tlm;
}

void traffic_light_manager::init_light_from_db(int lid)
{
    std::string sql = "select light_id, lights_group_id, ip, x, y ,z, reader_id, state, port, physics_light_id, physics_light_direction, special_flag from dat_light";
    if(0 == lid){
        sql += ";";
    }else{
        sql += " where light_id = " + std::to_string(lid);
        sql += ";";
    }

    std::string err = "";

    YADB::CDBResultSet res;
    sDBConnPool.Query(sql.c_str(), res, err);
    int count =  res.GetRecordCount(err);
    if(count < 1){
        log_error("增加或修改失败，error：%s", sql.c_str());
        return;
    }

    while(res.GetNextRecod(err)){
        traffic_light_ptr pl = std::make_shared<traffic_light>();
        if(nullptr == pl){
            return;
        }

        res.GetField("light_id", pl->m_light_id, err);
        res.GetField("lights_group_id", pl->m_group_id, err);
        res.GetField("ip", pl->m_ip, err);
        res.GetField("x", pl->x, err);
        res.GetField("y", pl->y, err);
        res.GetField("z", pl->z, err);
        res.GetField("reader_id", pl->m_site_id, err);
        res.GetField("state", pl->m_state, err);
        res.GetField("port", pl->m_port, err);
        res.GetField("physics_light_id", pl->m_phy_light_id, err);
        res.GetField("physics_light_direction", pl->m_phy_direction, err);
        res.GetField("special_flag", pl->m_special, err);
        m_unmap_lights[pl->m_light_id] = pl;
    }

    logn_info(2, "sql:%s",sql.c_str());
}

void traffic_light_manager::init_light_group_from_db(int gid)
{
    std::string sql = "select lights_group_id, x, y, z, scope, map_id, area_id, manual_control_time, light_auto_interval from dat_lights_group";
    if(gid > 0){
        sql += " where lights_group_id = " + std::to_string(gid);
    }
    sql += ";";

    std::string err = "";
    YADB::CDBResultSet res;
    
    sDBConnPool.Query(sql.c_str(), res, err);
    int count = res.GetRecordCount(err);
    if(count < 1){
        log_error("增加或修改失败，数据库中找不到：sql=%s", sql.c_str());
        return;
    }

    while(res.GetNextRecod(err)){
        traffic_light_group_ptr pg = std::make_shared<traffic_light_group>();
        if(nullptr == pg){
            return;
        }

        res.GetField("lights_group_id", pg->m_group_id, err);
        res.GetField("x", pg->x, err);
        res.GetField("y", pg->y, err);
        res.GetField("z", pg->z, err);
        res.GetField("scope", pg->m_scope, err);
        res.GetField("map_id", pg->m_map_id, err);
        res.GetField("area_id", pg->m_area_id, err);
        int ai = 0;
        res.GetField("light_auto_interval", ai, err);
        pg->m_auto_interval = ai;
        m_unmap_groups[pg->m_group_id] = pg;
    }
    
    logn_info(2, "sql:%s",sql.c_str());
}

void traffic_light_manager::init(const traffic_send_callback& tcb)
{
    init_light_from_db();
    init_light_group_from_db();

    for(hashmap_light::iterator it_light = m_unmap_lights.begin(); it_light != m_unmap_lights.end(); ++it_light){
        if(it_light->second->m_group_id >0){
            hashmap_group::iterator it_group = m_unmap_groups.find(it_light->second->m_group_id);
            if(it_group != m_unmap_groups.end()){
                it_group->second->insert(it_light->second);
            }
        }
    }

    set_send_callback(tcb);
}

void traffic_light_manager::start()
{
    log_info("[traffic_light] start traffic light service");
    m_stop = false;
    m_crossing_rule = std::unique_ptr<crossing_rule>(new crossing_rule);
    if(nullptr == m_crossing_rule){
        log_info("[traffic_light] create crossing rule failed");
        return;
    }

    m_avoidance_rule = std::unique_ptr<avoidance_rule>(new avoidance_rule);
    if(nullptr == m_avoidance_rule){
        log_info("[traffic_light] create avoidance rule failed");
        return;
    }

    m_thread_light.reset(new std::thread(std::bind(&traffic_light_manager::run, this)));
}

void traffic_light_manager::stop()
{
    log_info("[traffic_light] stop traffic light service");
    m_stop = true;
    m_thread_light->join();
}

void traffic_light_manager::put(const light_message& msg)
{
    std::unique_lock<std::mutex> lock(m_mutex);
    m_list_data.push_back(msg);
    m_condition.notify_one();
}

void traffic_light_manager::run()
{
    while(!m_stop){
        std::list<light_message> tmp_data;
        tmp_data.clear();

        {
            std::unique_lock<std::mutex> lock(m_mutex);
            while(m_list_data.empty()){
                m_condition.wait(lock);
            }

            if(m_list_data.size() > 0){
                m_list_data.swap(tmp_data);
            }
        }

        for(std::list<light_message>::iterator it = tmp_data.begin(); it != tmp_data.end(); ++it){
            switch(it->m_cmd){
                case cmd_reload:
                    log_info("[traffic_light] cmd=%d, gid=%d, lid=%d",it->m_cmd, it->m_group_id, it->m_light_id);
                    handle_reload(it->m_group_id, it->m_light_id);
                    break;
                case cmd_manual_ctrl:
                    log_info("[traffic_light] cmd=%d, gid=%d, lid=%d, name=%s, lc=%d", it->m_cmd, it->m_group_id, it->m_light_id, it->m_ctrl_name.c_str(), it->m_light_status);
                    handle_manual(it->m_group_id,it->m_light_id, it->m_ctrl_name, it->m_light_status);
                    break;
                case cmd_card_data:
                    log_info("[traffic_light] cmd=%d, cid=%d, bigger=%d, aid=%d, speed=%.2f", it->m_cmd, it->m_pos.m_card_id, it->m_pos.m_bigger, it->m_pos.m_area_id, it->m_pos.m_speed);
                    handle_position(it->m_pos);
                    break;
                default:
                    log_warn("[traffic_light] message cmd error, cmd=%d", it->m_cmd);
                    break;
            }
        
        }
    }
}

void traffic_light_manager::handle_reload(const int& gid, const int& lid)
{
    if(lid > 0){
        init_light_from_db(lid);
    } 
    if(gid > 0){
        init_light_group_from_db(gid);
    }
}

int traffic_light_manager::reload_light(const int& lid)
{
    init_light_from_db(lid);
    return 0;
}

int traffic_light_manager::reload_group(const int& gid)
{
    init_light_group_from_db(gid);
    return 0;
}

void traffic_light_manager::handle_manual(const int& gid, const int& ld, const std::string& name, const int& lc)
{
    traffic_light_group_ptr pg = find_group(gid);
    if(nullptr != pg){
        pg->set_manual_ctrl(name, ld, lc);
    }
}

void traffic_light_manager::handle_position(pos_data& p)
{
    if(p.m_type == 2){
        std::map<uint64_t, pos_data>::iterator it = m_map_card.find(p.m_card_id);
        if(it != m_map_card.end()){
            m_map_card.erase(it);
        }
    }else if(p.m_type == 3 || p.m_type == 1){
        m_geo_list.update(p.x, p.y, p.m_card_id);
        m_map_card[p.m_card_id] = p;
        // 路口规则
        if(nullptr != m_crossing_rule){
            m_crossing_rule->handle_rule(p);
        }

        if(!p.m_bigger && nullptr != m_avoidance_rule){
            m_avoidance_rule->handle_rule(p);
        }
    }
}
